from typing import Any, Dict

import os
import json


WORKSPACE = os.getcwd()

def load_config() -> Dict[str, Any]:
    with open(os.path.join(WORKSPACE, 'extr-config.json'), 'r', encoding='utf-8') as config_file:
        return json.loads(config_file.read())

def save_config(config: Dict[str, Any]) -> None:
    with open(os.path.join(WORKSPACE, 'extr-config.json'), 'w', encoding='utf-8') as config_file:
        config_file.write(
            json.dumps(config, indent=2)
        )

def create_source_file():
    config = load_config()
    file_path = os.path.join(WORKSPACE, '1', config['source'])
    with open(file_path, 'w', encoding='utf-8') as source_file:
        source_file.write('')

def create_labels_file():
    src = """from typing import Dict, List, Tuple
import re

from extr.regexes import RegEx, RegExLabel
from extr.relations import RegExRelationLabelBuilder


## ENTITIES
kb: Dict[str, List[str]]= {}
entity_patterns: List[RegExLabel] = []

## RELATIONS
relation_patterns: List[RegExLabel] = []

## ie. ('PERSON', 'ORG', 'NO_RELATION')
relation_defaults: List[Tuple[str, str, str]] = []
"""

    file_path = os.path.join(WORKSPACE, 'labels.py')
    with open(file_path, 'w', encoding='utf-8') as labels_file:
        labels_file.write(src)

def create_utils_file():
    src = """from typing import List


def sentence_tokenizer(text: str) -> List[List[str]]:
    return [
        text.split(' ')
    ]

def transform_text(text: str) -> str:
    return text
"""

    file_path = os.path.join(WORKSPACE, 'utils.py')
    with open(file_path, 'w', encoding='utf-8') as utils_file:
        utils_file.write(src)

def init() -> None:
    save_config({
        'source': 'source.txt',
        'split': {
            'amount': 25
        },
        'annotations': {
            'enable-html': True,
            'filter-redactions': True,
        }
    })

    directories = ['1', '2', '3', '4']

    for folder in directories:
        directory = os.path.join(WORKSPACE, folder)
        if not os.path.exists(directory):
            os.mkdir(directory)

    create_source_file()
    create_labels_file()
    create_utils_file()

def clean() -> None:
    files = [
        os.path.join(WORKSPACE, '4', 'ents.txt'),
        os.path.join(WORKSPACE, '4', 'ents-redacted.txt'),
        os.path.join(WORKSPACE, '4', 'rels.txt'),
    ]

    for file in files:
        if os.path.exists(file):
            os.remove(file)



from typing import Callable, Dict, List
import re
import os
import json

from extr.entities import create_entity_extractor
from extr.relations import RelationExtractor, HtmlRelationAnnotator
from extr_ds.models import RelationLabel
from extr_ds.labelers import RelationClassification

from .workspace import WORKSPACE
from ..utils.filesystem import load_data
from ..utils import imports


def get_labeler(sentence_tokenizer: Callable[[str], List[List[str]]]) -> RelationClassification:
    labels = imports.load_file(
        'labels',
        os.path.join(WORKSPACE, 'labels.py')
    )

    entity_extractor = create_entity_extractor(labels.entity_patterns, labels.kb)
    relation_extractor = RelationExtractor(labels.relation_patterns)

    return RelationClassification(
        sentence_tokenizer,
        entity_extractor,
        relation_extractor,
        labels.relation_defaults,
    )

def relate() -> None:
    utils = imports.load_file(
        'utils',
        os.path.join(WORKSPACE, 'utils.py')
    )

    labeler = get_labeler(utils.sentence_tokenizer)

    relation_groups: Dict[str, List[RelationLabel]] = {}
    for row in load_data(os.path.join(WORKSPACE, '2', 'dev.txt')):
        text = utils.transform_text(row)
        for relation_label in labeler.label(text):
            if not relation_label.definition in relation_groups:
                relation_groups[relation_label.definition] = []

            relation_groups[relation_label.definition].append(relation_label)

    rows = []
    blobs = []

    index = 0
    html_annotator = HtmlRelationAnnotator()
    for key, items in relation_groups.items():
        rows.append(f'<tr><td class="header" colspan=3>{key}</td></tr>')
        for relation_label in items:
            text = html_annotator.annotate(
                re.sub(r'</?e\d+>', '', relation_label.sentence),
                relation_label.relation
            )

            row_id = str(index)
            rows.append(f'<tr id="{row_id}"><td>{row_id}</td><td class="label">{relation_label.relation.label}</td><td>{text}</td></tr>')
            blobs.append(relation_label.todict())

            index += 1

    styles = """
span.entity { border: 1px solid black; border-radius: 5px; padding: 5px; margin: 3px; cursor: pointer; }
span.label { font-weight: bold; padding: 3px; color: black; }
span.e1 { background-color: aqua; }
span.e2 { background-color: coral; }
tr.delete { background-color: black !important; }
tr.delete span { background-color: black !important; }
td { line-height: 30px; border: 1px solid black; padding: 5px; }
td.header { font-weight: bold; }
td.label { font-weight: bold; text-align: center; }
"""

    html = """
<html>
    <head>
        <style>""" + styles + """</style>
    </head>
    <body><table>""" + '\n'.join(rows) + """</table></body>
</html>
"""

    html_path = os.path.join(WORKSPACE, '3', 'dev-rels.html')
    with open(html_path, 'w', encoding='utf-8') as html_file:
        html_file.write(html)

    with open(os.path.join(WORKSPACE, '3', 'dev-rels.json'), 'w', encoding='utf-8') as relation_outputs:
        relation_outputs.write(json.dumps(blobs, indent=2))

def change_label(label: str, row: int) -> None:
    html_path = os.path.join(WORKSPACE, '3', 'dev-rels.html')
    with open(html_path, 'r', encoding='utf-8') as html_file:
        html = html_file.read()

    html = re.sub(
        r'(<tr id="' + str(row) + '"><td>' + str(row) + '</td><td class="label">)(.+?)(</td>)',
        r'\1' + label + r'\3',
        html
    )

    with open(html_path, 'w', encoding='utf-8') as html_file:
        html_file.write(html)

    dev_path = os.path.join(WORKSPACE, '3', 'dev-rels.json')
    with open(dev_path, 'r', encoding='utf-8') as relation_outputs:
        dev = json.loads(relation_outputs.read())

    dev[row] = {
        'sentence': dev[row]['sentence'],
        'label': label
    }

    with open(os.path.join(WORKSPACE, '3', 'dev-rels.json'), 'w', encoding='utf-8') as relation_outputs:
        relation_outputs.write(json.dumps(dev, indent=2))

def delete_row(row: int) -> None:
    html_path = os.path.join(WORKSPACE, '3', 'dev-rels.html')
    with open(html_path, 'r', encoding='utf-8') as html_file:
        html = html_file.read()

    html = re.sub(
        r'(<tr )(id="' + str(row) + '")',
        r'\1class="delete" \2',
        html
    )

    with open(html_path, 'w', encoding='utf-8') as html_file:
        html_file.write(html)

    dev_path = os.path.join(WORKSPACE, '3', 'dev-rels.json')
    with open(dev_path, 'r', encoding='utf-8') as relation_outputs:
        dev = json.loads(relation_outputs.read())

    dev[row] = {
        'sentence': dev[row]['sentence'],
        'label': dev[row]['label'],
        'attribute': 'delete',
    }

    with open(os.path.join(WORKSPACE, '3', 'dev-rels.json'), 'w', encoding='utf-8') as relation_outputs:
        relation_outputs.write(json.dumps(dev, indent=2))



import sys
from typing import List, Optional, Tuple, cast
from dataclasses import dataclass

from .process import workspace, entities, relations
from .process import split, \
                     save_entities, \
                     save_relations


@dataclass
class Command:
    branch: str
    sub_command: Optional[str]
    args: Optional[str]

    @property
    def has_sub_command(self) -> bool:
        return not self.sub_command is None

    @property
    def has_args(self) -> bool:
        return not self.args is None

    def get_args(self, delim='=') -> Optional[Tuple[str, List[str]]]:
        if self.has_args:
            a1, a2 = cast(str, self.args).split(delim)
            return (a1, a2.split(','))

        return None

def create_command(args: List[str]) -> Command:
    return Command(
        branch=args[0],
        sub_command=args[1] if len(args) > 1 else None,
        args=args[2] if len(args) > 2 else None
    )

def main() -> int:
    if len(sys.argv) == 1:
        return -1

    command = create_command(sys.argv[1:])

    if command.branch == '--init':
        workspace.init()

    if command.branch == '--reset':
        workspace.clean()

    if command.branch == '--split':
        split()
        entities.annotate()
        relations.relate()

    if command.branch == '--annotate':
        if not command.has_sub_command:
            entities.annotate()
            relations.relate()
            return 0

        if command.sub_command == '-ents':
            ## --annotate -ents
            entities.annotate()

        if command.sub_command == '-rels':
            ## --annotate -rels
            relations.relate()

    if command.branch == '--relate':
        if not command.has_sub_command:
            return -1

        if command.sub_command == '-label':
            ## --relate -label NO_RELATION=1,3,6
            label, args = cast(Tuple[str, List[str]], command.get_args())
            for index in map(int, args):
                relations.change_label(label, index)

        if command.sub_command == '-delete':
            ## --relate -delete 1,3,6
            for index in map(int, cast(str, command.args).split(',')):
                relations.delete_row(index)

    if command.branch == '--save':
        if not command.has_sub_command:
            save_entities()
            save_relations()
            return 0

        if command.sub_command == '-ents':
            ## --save -ents
            save_entities()

        if command.sub_command == '-rels':
            ## --save -relts
            save_relations()

    return 0



from typing import List, Dict
from dataclasses import dataclass
from extr.models import Token, Relation


@dataclass
class Label:
    tokens: List[Token]
    labels: List[str]

    def __repr__(self) -> str:
        return f'<Label tokens={self.tokens}, labels={self.labels}>'

@dataclass
class RelationLabel:
    sentence: str
    relation: Relation

    @property
    def label(self) -> str:
        return self.relation.label

    @property
    def definition(self) -> str:
        return self.relation.definition

    def todict(self) -> Dict[str, str]:
        return {
            'sentence': self.sentence,
            'label': self.label,
        }

    def __repr__(self) -> str:
        return f'<RelationLabel sentence="{self.sentence}" label="{self.label}">'



from typing import Any, Dict
import os
import json

from .workspace import WORKSPACE
from ..utils.filesystem import load_data, \
                               append_data


def save_entities() -> None:
    dataset = set(load_data(os.path.join(WORKSPACE, '2', 'dev.txt')))
    append_data(os.path.join(WORKSPACE, '4', 'ents.txt'), dataset)

    redacted_dataset = set(load_data(os.path.join(WORKSPACE, '3', 'dev-ents-redacted.txt')))
    append_data(os.path.join(WORKSPACE, '4', 'ents-redacted.txt'), redacted_dataset)

def save_relations() -> None:
    with open(os.path.join(WORKSPACE, '3', 'dev-rels.json'), 'r', encoding='utf-8') as relation_outputs:
        data = [row for row in json.loads(relation_outputs.read()) if not 'attribute' in row]

    output_path = os.path.join(WORKSPACE, '4', 'rels.json')
    if os.path.exists(output_path):
        with open(output_path, 'r', encoding='utf-8') as relation_outputs:
            current_data = json.loads(relation_outputs.read())

        data.extend(current_data)

    lookup: Dict[str, Dict[str, Any]] = {}
    for row in data:
        key = row['sentence'].lower()
        if not key in lookup:
            lookup[key] = { 'labels': set(), 'items': [] }

        lookup[key]['labels'].add(row['label'])
        lookup[key]['items'].append(row)

    slim_data = []
    for key, item in lookup.items():
        labels = item['labels']
        items = item['items']

        if len(labels) == 1:
            slim_data.append(items[0])
        else:
            print('Label mismatch found (skipping):')
            print('sentence:', items[0]['sentence'])
            print('labels:', ', '.join(labels))
            print()

    with open(output_path, 'w', encoding='utf-8') as relation_outputs:
        relation_outputs.write(json.dumps(slim_data, indent=2))



from typing import Dict, List, Set, cast
import os
import re
import json
from dataclasses import dataclass, field

from extr.entities import create_entity_extractor, \
                          LabelOnlyEntityAnnotator, \
                          HtmlEntityAnnotator, \
                          EntityExtractor

from .workspace import load_config, WORKSPACE
from ..utils.filesystem import load_data, save_data
from ..utils import imports


entity_annotator = LabelOnlyEntityAnnotator()
entity_html_annotator = HtmlEntityAnnotator()

@dataclass()
class Annotations:
    text: List[str] = field(default_factory=lambda: [])
    html: List[str] = field(default_factory=lambda: [])
    text_by_label: Dict[str, List[str]] = field( default_factory=lambda: {})

def get_extractor() -> EntityExtractor:
    labels = imports.load_file(
        'labels',
        os.path.join(WORKSPACE, 'labels.py')
    )

    return create_entity_extractor(labels.entity_patterns, labels.kb)

def annotate_file(file_path: str) -> Annotations:
    entity_extractor = get_extractor()
    utils = imports.load_file('utils', os.path.join(WORKSPACE, 'utils.py'))

    cache = Annotations()
    text_by_label: Dict[str, Set[str]] = {}

    for row in load_data(file_path):
        text = cast(str, utils.transform_text(row))
        entities = entity_extractor.get_entities(text)

        cache.text.append(
            entity_annotator.annotate(text, entities).annotated_text
        )

        cache.html.append(
            entity_html_annotator.annotate(text, entities).annotated_text
        )

        for entity in entities:
            if not entity.label in text_by_label:
                text_by_label[entity.label] = set()

            text_by_label[entity.label].add(entity.text)

    for key, value in text_by_label.items():
        cache.text_by_label[key] = list(sorted(value))

    return cache

def create_redacted_file(annotations: List[str]) -> None:
    def get_redacted_templates() -> Set[str]:
        redacted_templates = set()
        redacted_templates_file_path = os.path.join(WORKSPACE, '4', 'ents-redacted.txt')
        if os.path.exists(redacted_templates_file_path):
            redacted_templates = set(load_data(redacted_templates_file_path))

        return redacted_templates

    config = load_config()
    annotations_config = config['annotations'] if 'annotations' in config else {}

    redactions: List[str] = [
        re.sub(' +', ' ', re.sub(r'<[A-Z]+>.+?</[A-Z]+>', ' ', row))
        for row in annotations
    ]

    if annotations_config['filter-redactions'] if 'filter-redactions' in annotations_config else True:
        redacted_templates = get_redacted_templates()
        redactions = [row for row in redactions if not row in redacted_templates]

    save_data(
        os.path.join(WORKSPACE, '3', 'dev-ents-redacted.txt'),
        redactions
    )

def create_parsed_by_file(text_by_label: Dict[str, List[str]]) -> None:
    stats_path = os.path.join(WORKSPACE, '3', 'dev-ents.stats.json')
    with open(stats_path, 'w', encoding='UTF8') as dev_stats:
        dev_stats.write(json.dumps(text_by_label, indent=2))

def create_html_file(annotations: List[str]) -> None:
    styles = """
p { margin: 5px; line-height: 45px; }
span.entity { border: 1px solid black; border-radius: 5px; padding: 5px; margin: 3px; cursor: pointer; }
span.label { font-weight: bold; padding: 3px; color: black; }
"""

    custom_styles_path = os.path.join(WORKSPACE, 'styles.css')
    if os.path.exists(custom_styles_path):
        with open(custom_styles_path, 'r', encoding='utf-8') as custom_styles:
            styles += custom_styles.read()

    rows = '<hr />\n'.join(
        [f'<p>{annotation}</p>' for annotation in annotations]
    )

    html = """
<html>
    <head>
        <style>
            """ + \
            styles + \
            """
        </style>
    </head>
    <body>""" + rows + """</body>
</html>
"""
    html_path = os.path.join(WORKSPACE, '3', 'dev-ents.html')
    with open(html_path, 'w', encoding='utf-8') as html_file:
        html_file.write(html)

def annotate() -> None:
    cache = annotate_file(os.path.join(WORKSPACE, '2', 'dev.txt'))

    save_data(
        os.path.join(WORKSPACE, '3', 'dev-ents.txt'),
        cache.text
    )

    create_redacted_file(cache.text)
    create_parsed_by_file(cache.text_by_label)

    config = load_config()
    annotations_config = config['annotations'] if 'annotations' in config else {}
    if annotations_config['enable-html'] if 'enable-html' in annotations_config else True:
        create_html_file(cache.html)
